(function() {
  var Codec, PATH_SEP, SEP, SUBKEY_SEP, SUBKEY_SEPS, SubkeyCodec, UNSAFE_CHARS, _escapeString, consts, defineProperty, isFunction, isString, path, register, relativePath, resolvePathArray, toPath, toPathArray;

  Codec = require('buffer-codec');

  SEP = require('./separator');

  path = require('../path');

  consts = require('../consts');

  isString = require("abstract-object/lib/util/isString");

  isFunction = require("abstract-object/lib/util/isFunction");

  register = Codec.register;

  _escapeString = Codec.escapeString;

  toPath = path.join;

  toPathArray = path.toArray;

  relativePath = path.relative;

  resolvePathArray = path.resolveArray;

  SUBKEY_SEPS = SEP.SUBKEY_SEPS;

  UNSAFE_CHARS = SEP.UNSAFE_CHARS;

  PATH_SEP = SUBKEY_SEPS[0][0];

  SUBKEY_SEP = SUBKEY_SEPS[1][0];

  defineProperty = Object.defineProperty;

  module.exports = SubkeyCodec = (function() {
    var DEL_OP, GET_OP, HALT_OP, PUT_OP, SKIP_OP, TRANS_OP, _decode, _encode, _encodeKey, addEncodings, decodeKey, encode, encodeKey, escapeString, getPathArray, indexOfType, prepareKeyPath, prepareOperation, resolveKeyPath, unescapeString;

    function SubkeyCodec() {}

    GET_OP = consts.GET_OP;

    PUT_OP = consts.PUT_OP;

    DEL_OP = consts.DEL_OP;

    TRANS_OP = consts.TRANS_OP;

    HALT_OP = consts.HALT_OP;

    SKIP_OP = consts.SKIP_OP;

    defineProperty(SubkeyCodec, "PATH_SEP", {
      get: function() {
        return PATH_SEP;
      }
    });

    defineProperty(SubkeyCodec, "SUBKEY_SEP", {
      get: function() {
        return SUBKEY_SEP;
      }
    });

    defineProperty(SubkeyCodec, "SUBKEY_SEPS", {
      get: function() {
        return SUBKEY_SEPS;
      },
      set: function(value) {
        SEP.SUBKEY_SEPS = value;
        SUBKEY_SEPS = SEP.SUBKEY_SEPS;
        UNSAFE_CHARS = SEP.UNSAFE_CHARS;
        PATH_SEP = SUBKEY_SEPS[0][0];
        return SUBKEY_SEP = SUBKEY_SEPS[1][0];
      }
    });

    addEncodings = function(op, aParent) {
      if (aParent && aParent._options) {
        op.keyEncoding || (op.keyEncoding = aParent._options.keyEncoding);
        op.valueEncoding || (op.valueEncoding = aParent._options.valueEncoding);
      }
      return op;
    };

    SubkeyCodec.lowerBound = '\u0000';

    SubkeyCodec.upperBound = '\udbff\udfff';

    SubkeyCodec.escapeString = escapeString = function(aString, aUnSafeChars) {
      if (aUnSafeChars == null) {
        aUnSafeChars = UNSAFE_CHARS;
      }
      return _escapeString(aString, aUnSafeChars);
    };

    SubkeyCodec.unescapeString = unescapeString = decodeURIComponent;

    SubkeyCodec.getPathArray = getPathArray = function(aPath, aRootPath) {
      if (aPath == null) {
        if (aRootPath == null) {
          return aPath;
        }
        if (isFunction(aRootPath.pathAsArray)) {
          return aRootPath.pathAsArray();
        }
        return aRootPath;
      }
      if (isFunction(aPath.pathAsArray)) {
        return aPath.pathAsArray();
      }
      if (isString(aPath)) {
        if (aRootPath) {
          if (isFunction(aRootPath.pathAsArray)) {
            aRootPath = aRootPath.pathAsArray();
          }
          aPath = resolvePathArray(aRootPath, aPath);
          aPath.shift(0, 1);
        } else {
          aPath = toPathArray(aPath);
        }
      }
      return aPath;
    };

    SubkeyCodec.prepareKeyPath = prepareKeyPath = function(aPathArray, aKey, op) {
      var i, vKey;
      vKey = aKey;
      if (isString(aKey) && aKey.length) {
        aPathArray = resolvePathArray(aPathArray, aKey);
        aPathArray.shift(0, 1);
        aKey = aPathArray.pop();
        vKey = aKey;
        i = SUBKEY_SEPS[0].indexOf(aKey[0], 1);
        if (i > 0) {
          if (!op.separator) {
            op.separator = aKey[0];
          }
          if (op.separator === aKey[0]) {
            aKey = aKey.slice(1);
          }
        }

        /*
         * remove separator parameter and add it to string:
        if op.separator && op.separator != PATH_SEP
          aKey = op.separator + aKey if aKey[0] != op.separator
          op.separator = undefined if aKey[0] is op.separator
         */
      }
      op._keyPath = [aPathArray, vKey];
      op.path = aPathArray;
      op.key = aKey;
    };

    SubkeyCodec.prepareOperation = prepareOperation = function(preHooks, operationType, aOperation, aPathArray, aKey) {
      var addOp, result, triggerArgs;
      if (aOperation.path) {
        addEncodings(aOperation, aOperation.path);
      }
      if (!aPathArray) {
        aPathArray = aOperation.path;
      }
      if (!aKey) {
        aKey = aOperation.key;
      }
      prepareKeyPath(aPathArray, aKey, aOperation);
      if (!aOperation.separator) {
        delete aOperation.separator;
      }
      if (preHooks && operationType && aOperation.triggerBefore !== false) {
        switch (operationType) {
          case PUT_OP:
          case DEL_OP:
            triggerArgs = [operationType, aOperation];
            break;
          case TRANS_OP:
            addOp = function(aOperation) {
              if (aOperation) {
                ops.push(aOperation);
              }
            };
            triggerArgs = [operationType, aOperation, addOp];
        }
        if (triggerArgs) {
          result = preHooks.trigger(operationType, triggerArgs);
        }
      }
      return result;
    };

    SubkeyCodec.resolveKeyPath = resolveKeyPath = function(aPathArray, aKey, op) {
      if (!op) {
        op = {};
      }
      prepareKeyPath(aPathArray, aKey, op);
      return [op.path, op.key];
    };

    SubkeyCodec.encode = encode = function(aPath, aKey, aSeperator, dontEscapeSeperator, keyEncoding) {
      var hasSep, i, keyIsStr;
      keyIsStr = isString(aKey) && aKey.length > 0;
      hasSep = !!aSeperator;
      if (!aSeperator) {
        aSeperator = SUBKEY_SEP;
      }
      if (keyIsStr) {
        if (hasSep && aKey[0] === aSeperator) {
          aKey = aKey.substring(1);
        }
      }
      if (dontEscapeSeperator === true) {
        if (hasSep && aSeperator !== PATH_SEP) {
          aSeperator = PATH_SEP + aSeperator;
        }
        hasSep = false;
      }
      if (hasSep) {
        i = SUBKEY_SEPS[0].indexOf(aSeperator);
        if (i >= 0) {
          aSeperator = SUBKEY_SEPS[1][i];
          if (aSeperator !== PATH_SEP) {
            aSeperator = PATH_SEP + SUBKEY_SEPS[1][i];
          }
        } else {
          aSeperator = PATH_SEP + aSeperator;
        }
      } else if (keyIsStr) {
        i = SUBKEY_SEPS[0].indexOf(aKey[0], 1);
        if (i > 0) {
          aSeperator = PATH_SEP + SUBKEY_SEPS[1][i];
          aKey = aKey.substring(1);
        }
      }
      if (keyIsStr) {
        aKey = escapeString(aKey);
      }
      if (aPath.length) {
        aPath = path.join(aPath);
      } else if (aSeperator.length >= 2 && aSeperator[0] === PATH_SEP) {
        aPath = "";
      } else {
        aPath = PATH_SEP;
      }
      if (keyEncoding) {
        aKey = keyEncoding.encode(aKey);
      }
      return aPath + aSeperator + aKey;
    };

    indexOfType = function(s) {
      var c, i;
      i = s.length - 1;
      while (i > 0) {
        c = s[i];
        if (SUBKEY_SEPS[1].indexOf(c) >= 0) {
          return i;
        }
        --i;
      }
      return -1;
    };

    SubkeyCodec.decode = _decode = function(s, aSeparator) {
      var i, j, result, vKey, vSep;
      i = indexOfType(s);
      if (i >= 0) {
        vSep = s[i];
        if (vSep === SUBKEY_SEP) {
          vSep = PATH_SEP;
        } else {
          j = SUBKEY_SEPS[1].indexOf(vSep);
          vSep = PATH_SEP + SUBKEY_SEPS[0][j];
        }
        vKey = unescapeString(s.substring(i + 1));
        result = [s.substring(1, i).split(PATH_SEP).filter(Boolean).map(unescapeString), vKey, vSep];
        if (isString(aSeparator) && aSeparator !== s[i]) {
          result.push(s[i]);
        }
      }
      return result;
    };

    SubkeyCodec._encodeKey = _encodeKey = function(aPathArray, aKey, keyEncoding, options) {
      var vSep, vSepRaw;
      if (options) {
        vSep = options.separator;
        vSepRaw = options.separatorRaw;
      }
      return encode(aPathArray, aKey, vSep, vSepRaw, keyEncoding);
    };

    SubkeyCodec.encodeKey = encodeKey = function(aPathArray, aKey, keyEncoding, options, operationType, preHooks) {
      prepareOperation(preHooks, operationType, options, aPathArray, aKey);
      aPathArray = options.path;
      aKey = options.key;
      return _encodeKey(aPathArray, aKey, keyEncoding, options);
    };

    SubkeyCodec.decodeKey = decodeKey = function(key, keyEncoding, options) {
      var v, vPath, vSep;
      v = _decode(key, options && options.separator);
      vSep = v[2];
      if (vSep == null) {
        vSep = PATH_SEP;
      }
      key = keyEncoding ? keyEncoding.decode(v[1], options) : v[1];
      if (options) {
        if (options.absoluteKey) {
          key = toPath(v[0]) + vSep + key;
        } else if (options.path && isString(key) && key !== "") {
          vPath = relativePath(options.path, v[0]);
          if (vPath === "" && vSep === PATH_SEP) {
            vSep = "";
          } else if (vSep.length >= 2) {
            vSep = vSep.substring(1);
          }
          key = vPath + vSep + key;
        }
      } else {
        key = v[1];
      }
      return key;
    };

    SubkeyCodec.encodeValue = function(value, options) {
      var encoding;
      if (encoding = Codec(options.valueEncoding)) {
        return encoding.encode(value);
      } else {
        return value;
      }
    };

    SubkeyCodec.decodeValue = function(value, options) {
      var encoding;
      if (encoding = Codec(options.valueEncoding)) {
        return encoding.decode(value);
      } else {
        return value;
      }
    };

    SubkeyCodec.applyEncoding = function(options) {
      if (options) {
        options.valueEncoding = Codec(options.valueEncoding);
        if (options.valueEncoding && options.valueEncoding.name === 'Text') {
          return options.valueEncoding = false;
        }
      }
    };

    SubkeyCodec._encode = _encode = function(e) {
      var dontEscapeSep, sep;
      if (e.length > 2) {
        sep = e[2];
      }
      if (e.length > 3) {
        dontEscapeSep = e[3];
      }
      return encode(e[0], e[1], sep, dontEscapeSep);

      /*
      vSeperator = SUBKEY_SEP
      #e[2]: seperator
      hasSep = e.length >= 3 && e[2]
      #e[3]: DontEscapeSep
      if e.length >=4 && e[3] is true
          hasSep = false
          if e[2] then vSeperator = e[2]
          if vSeperator isnt PATH_SEP then vSeperator = PATH_SEP + vSeperator
      key = e[1]
      isStrKey = isString(key) && key.length isnt 0
      if hasSep
        vSeperator = e[2]
        i = SUBKEY_SEPS[0].indexOf(vSeperator)
        if i >= 0
          vSeperator = SUBKEY_SEPS[1][i]
          if vSeperator isnt PATH_SEP then vSeperator = PATH_SEP + SUBKEY_SEPS[1][i]
        else
          vSeperator = PATH_SEP + vSeperator
      else if isStrKey
        #try to find the separator on the key
        i = SUBKEY_SEPS[0].indexOf(key[0], 1)
        if i > 0
            vSeperator = PATH_SEP + SUBKEY_SEPS[1][i]
            key = key.substring(1)
      if isStrKey
        if hasSep && key[0] == e[2] then key = key.substring(1)
        key = escapeString(key)
      #console.log("codec.encode:",path.join(e[0]) + vSeperator + key)
      #TODO: I should encode with path.join(e[0], vSeperator + key)) simply in V8.
       *      all separators are same now.
      vPath = PATH_SEP
      if e[0].length
        vPath = path.join(e[0])
      else if vSeperator.length >=2 && vSeperator[0] == PATH_SEP
        vPath = ""
      vPath + vSeperator + key
       */
    };

    return SubkeyCodec;

  })();

}).call(this);

//# sourceMappingURL=index.js.map
