(function() {
  var AbstractNoSQL, Codec, EncodingNoSQL, Errors, InvalidArgumentError, PATH_SEP, SUBKEY_SEP, SubkeyCache, SubkeyIterator, SubkeyNoSQL, WriteError, _encodeKey, addPathToRange, codec, consts, decodeKey, encodeKey, extend, getPathArray, hooks, inherits, inheritsDirectly, isArray, isFunction, isInheritedFrom, isObject, isString, lowerBound, path, prepareKeyPath, prepareOperation, relativePath, resolveKeyPath, resolvePathArray, subkey, toPath, upperBound;

  Errors = require("abstract-error");

  EncodingNoSQL = require("nosql-encoding");

  AbstractNoSQL = EncodingNoSQL.super_;

  SubkeyIterator = require("./subkey-iterator");

  SubkeyCache = require("./subkey-cache");

  Codec = require("buffer-codec");

  inherits = require("inherits-ex/lib/inherits");

  inheritsDirectly = require("inherits-ex/lib/inheritsDirectly");

  isInheritedFrom = require("inherits-ex/lib/isInheritedFrom");

  isFunction = require("util-ex/lib/is/type/function");

  isString = require("util-ex/lib/is/type/string");

  isObject = require("util-ex/lib/is/type/object");

  isArray = require("util-ex/lib/is/type/array");

  extend = require("util-ex/lib/_extend");

  hooks = require("./hooks");

  path = require("./path");

  codec = require("./codec");

  consts = require("./consts");

  addPathToRange = require("./range").addPathToRange;

  subkey = require("./subkey");

  PATH_SEP = codec.PATH_SEP;

  SUBKEY_SEP = codec.SUBKEY_SEP;

  lowerBound = codec.lowerBound;

  upperBound = codec.upperBound;

  resolveKeyPath = codec.resolveKeyPath;

  prepareKeyPath = codec.prepareKeyPath;

  prepareOperation = codec.prepareOperation;

  _encodeKey = codec._encodeKey;

  encodeKey = codec.encodeKey;

  decodeKey = codec.decodeKey;

  getPathArray = codec.getPathArray;

  InvalidArgumentError = Errors.InvalidArgumentError;

  WriteError = Errors.WriteError;

  toPath = path.join;

  relativePath = path.relative;

  resolvePathArray = path.resolveArray;

  module.exports = SubkeyNoSQL = (function() {
    var DEL_OP, GET_OP, HALT_OP, PUT_OP, SKIP_OP, TRANS_OP, eventable;

    inherits(SubkeyNoSQL, EncodingNoSQL);

    SubkeyNoSQL.GET_OP = GET_OP = consts.GET_OP;

    SubkeyNoSQL.PUT_OP = PUT_OP = consts.PUT_OP;

    SubkeyNoSQL.DEL_OP = DEL_OP = consts.DEL_OP;

    SubkeyNoSQL.TRANS_OP = TRANS_OP = consts.TRANS_OP;

    SubkeyNoSQL.HALT_OP = HALT_OP = consts.HALT_OP;

    SubkeyNoSQL.SKIP_OP = SKIP_OP = consts.SKIP_OP;

    function SubkeyNoSQL(aClass) {
      var vIt, vIteratorClass, vParentClass;
      if (!(this instanceof SubkeyNoSQL)) {
        if (isInheritedFrom(aClass, 'SubkeyNoSQL')) {
          console.error(aClass.name + " class has already been subkey-able.");
          return aClass;
        }
        if (EncodingNoSQL) {
          vParentClass = isInheritedFrom(aClass, EncodingNoSQL);
        }
        if (!vParentClass) {
          vParentClass = isInheritedFrom(aClass, 'AbstractNoSQL');
        }
        if (vParentClass) {
          inheritsDirectly(vParentClass, SubkeyNoSQL);
          vIteratorClass = aClass.prototype.IteratorClass;
          if (vIteratorClass && !isInheritedFrom(vIteratorClass, 'SubkeyIterator')) {
            if (vIt = isInheritedFrom(vIteratorClass, 'EncodingIterator')) {
              vIteratorClass = vIt;
            } else {
              vIteratorClass = isInheritedFrom(vIteratorClass, 'AbstractIterator');
            }
            if (vIteratorClass) {
              inheritsDirectly(vIteratorClass, SubkeyIterator);
            }
          }
          return aClass;
        } else {
          throw new InvalidArgumentError("class should be inherited from EncodingNoSQL or AbstractNoSQL");
        }
      }
      AbstractNoSQL.prototype.constructor.apply(this, arguments);
    }

    SubkeyNoSQL.prototype.initialize = function() {
      this.cache = new SubkeyCache();
      this.SubkeyClass = subkey(this);
      return SubkeyNoSQL.__super__.initialize.apply(this, arguments);
    };

    SubkeyNoSQL.prototype.finalize = function() {
      this.cache.free();
      this.cache = void 0;
      this.SubkeyClass = void 0;
      return SubkeyNoSQL.__super__.finalize.apply(this, arguments);
    };


    /*
      first check preHooks if operationType
      return false if prehooks disallow this operation
      or return encoded string.
     */

    SubkeyNoSQL.prototype.encodeKey = function(aPathArray, aKey, op, operationType) {
      var keyEncoding, result;
      result = prepareOperation(this.preHooks, operationType, op, aPathArray, aKey);
      if (result === HALT_OP) {
        return false;
      }
      keyEncoding = this.keyEncoding(op);
      if (result === SKIP_OP) {
        if (keyEncoding) {
          aKey = keyEncoding.encode(aKey);
        }
        return aKey;
      }
      return _encodeKey(op.path, op.key, keyEncoding, op);
    };

    SubkeyNoSQL.prototype.decodeKey = function(key, options) {
      var keyEncoding;
      keyEncoding = this.keyEncoding(options);
      return decodeKey(key, keyEncoding, options);
    };

    SubkeyNoSQL.prototype.setOpened = function(isOpened, options) {
      if (isOpened) {
        this.preHooks = hooks();
        this.postHooks = hooks();
        this.cache.reset(options);
        if (options && options.path) {
          this._pathArray = getPathArray(options.path);
          delete options.path;
        } else {
          this._pathArray = [];
        }
      } else {
        if (this.preHooks) {
          this.preHooks.free();
        }
        if (this.postHooks) {
          this.postHooks.free();
        }
        this.cache.clear();
        this.preHooks = null;
        this.postHooks = null;
        this._pathArray = null;
      }
      return SubkeyNoSQL.__super__.setOpened.call(this, isOpened, options);
    };

    SubkeyNoSQL.prototype.getPathArray = function(options, aParentPath) {
      var result, vRootPath;
      vRootPath = this._pathArray;
      if (aParentPath) {
        vRootPath = getPathArray(aParentPath, vRootPath);
      }
      result = options && options.path ? getPathArray(options.path, vRootPath) : vRootPath;
      return result;
    };

    SubkeyNoSQL.prototype.isExistsSync = function(key, options) {
      path = this.getPathArray(options);
      options = extend({}, this._options, options);
      key = this.encodeKey(path, key, options);
      if (key === false) {
        return false;
      }
      return AbstractNoSQL.prototype.isExistsSync.call(this, key, options);
    };

    SubkeyNoSQL.prototype.isExistsAsync = function(key, options, callback) {
      path = this.getPathArray(options);
      if (isFunction(options)) {
        callback = options;
        options = void 0;
      }
      options = extend({}, this._options, options);
      key = this.encodeKey(path, key, options);
      if (key === false) {
        return false;
      }
      return AbstractNoSQL.prototype.isExistsAsync.call(this, key, options, callback);
    };

    SubkeyNoSQL.prototype.getSync = function(key, options) {
      var encoding, result;
      path = this.getPathArray(options);
      options = extend({}, this._options, options);
      key = this.encodeKey(path, key, options, GET_OP);
      if (key === false) {
        this.dispatchError(new InvalidArgumentError('getSync:can not encodeKey'));
        return;
      }
      result = AbstractNoSQL.prototype.getSync.call(this, key, options);
      if (result !== void 0) {
        encoding = this.valueEncoding(options);
        if (encoding) {
          result = encoding.decode(result);
        }
      }
      return result;
    };

    SubkeyNoSQL.prototype.getAsync = function(key, options, callback) {
      var encoding;
      path = this.getPathArray(options);
      if (isFunction(options)) {
        callback = options;
        options = void 0;
      }
      options = extend({}, this._options, options);
      key = this.encodeKey(path, key, options, GET_OP);
      if (key === false) {
        this.dispatchError(new InvalidArgumentError('getAsync:can not encodeKey'), callback);
        return;
      }
      encoding = this.valueEncoding(options);
      return AbstractNoSQL.prototype.getAsync.call(this, key, options, (function(_this) {
        return function(err, value) {
          if (err) {
            return _this.dispatchError(err, callback);
          }
          if (encoding) {
            value = encoding.decode(value);
          }
          return callback(null, value);
        };
      })(this));
    };

    SubkeyNoSQL.prototype.getBufferSync = function(key, destBuffer, options) {
      path = this.getPathArray(options);
      options = extend({}, this._options, options);
      key = this.encodeKey(path, key, options, GET_OP);
      if (key === false) {
        return false;
      }
      return AbstractNoSQL.prototype.getBufferSync.call(this, key, destBuffer, options);
    };

    SubkeyNoSQL.prototype.getBufferAsync = function(key, destBuffer, options, callback) {
      path = this.getPathArray(options);
      if (isFunction(options)) {
        callback = options;
        options = void 0;
      }
      options = extend({}, this._options, options);
      key = this.encodeKey(path, key, options, GET_OP);
      if (key === false) {
        return false;
      }
      return AbstractNoSQL.prototype.getBufferAsync.call(this, key, destBuffer, options, callback);
    };

    SubkeyNoSQL.prototype.mGetSync = function(keys, options) {
      var j, k, len, result, vOptions, valueEncoding;
      path = this.getPathArray(options);
      options = extend({}, this._options, options);
      valueEncoding = this.valueEncoding(options);
      result = [];
      vOptions = extend({}, options);
      for (j = 0, len = keys.length; j < len; j++) {
        k = keys[j];
        k = this.encodeKey(path, k, vOptions, GET_OP);
        vOptions.path = options.path;
        if (k !== false) {
          result.push(k);
        }
      }
      result = AbstractNoSQL.prototype.mGetSync.call(this, result, options);
      if (options.keys !== false) {
        result.map((function(_this) {
          return function(item) {
            item.key = _this.decodeKey(item.key, options);
            if (valueEncoding) {
              return item.value = valueEncoding.decode(item.value);
            }
          };
        })(this));
      } else {
        if (valueEncoding) {
          result = result.map(valueEncoding.decode.bind(valueEncoding));
        }
      }
      return result;
    };

    SubkeyNoSQL.prototype.mGetAsync = function(keys, options, callback) {
      var j, k, len, that, vKeys, vOptions, valueEncoding;
      path = this.getPathArray(options);
      if (isFunction(options)) {
        callback = options;
        options = void 0;
      }
      options = extend({}, this._options, options);
      valueEncoding = this.valueEncoding(options);
      vKeys = [];
      vOptions = extend({}, options);
      for (j = 0, len = keys.length; j < len; j++) {
        k = keys[j];
        k = this.encodeKey(path, k, vOptions, GET_OP);
        vOptions.path = options.path;
        if (k !== false) {
          vKeys.push(k);
        }
      }
      that = this;
      return AbstractNoSQL.prototype.mGetAsync.call(this, vKeys, options, function(err, result) {
        if (err) {
          return that.dispatchError(err, callback);
        }
        if (options.keys !== false) {
          result.map(function(item) {
            item.key = that.decodeKey(item.key, options);
            if (valueEncoding) {
              return item.value = valueEncoding.decode(item.value);
            }
          });
        } else {
          if (valueEncoding) {
            result = result.map(valueEncoding.decode.bind(valueEncoding));
          }
        }
        return callback(null, result);
      });
    };

    SubkeyNoSQL.prototype.putSync = function(key, value, options) {
      var result, vKeyPath, valueEncoding;
      path = this.getPathArray(options);
      options = extend({}, this._options, options);
      key = this.encodeKey(path, key, options, PUT_OP);
      if (key === false) {
        return false;
      }
      valueEncoding = this.valueEncoding(options);
      if (valueEncoding) {
        value = valueEncoding.encode(value);
      }
      result = AbstractNoSQL.prototype.putSync.call(this, key, value, options);
      if (result) {
        vKeyPath = options._keyPath;
        options.path = vKeyPath[0];
        options.key = vKeyPath[1];
        delete options._keyPath;
        this.postHooks.trigger(PUT_OP, vKeyPath, [PUT_OP, options]);
      }
      return result;
    };

    SubkeyNoSQL.prototype.putAsync = function(key, value, options, callback) {
      var valueEncoding;
      path = this.getPathArray(options);
      if (isFunction(options)) {
        callback = options;
        options = void 0;
      }
      options = extend({}, this._options, options);
      key = this.encodeKey(path, key, options, PUT_OP);
      if (key === false) {
        return false;
      }
      valueEncoding = this.valueEncoding(options);
      if (valueEncoding) {
        value = valueEncoding.encode(value);
      }
      return AbstractNoSQL.prototype.putAsync.call(this, key, value, options, (function(_this) {
        return function(err, result) {
          var vKeyPath;
          if (err) {
            return _this.dispatchError(err, callback);
          }
          vKeyPath = options._keyPath;
          options.path = vKeyPath[0];
          options.key = vKeyPath[1];
          delete options._keyPath;
          _this.postHooks.trigger(PUT_OP, vKeyPath, [PUT_OP, options]);
          if (callback) {
            return callback(null, result);
          }
        };
      })(this));
    };

    SubkeyNoSQL.prototype.delSync = function(key, options) {
      var result, vKeyPath;
      path = this.getPathArray(options);
      options = extend({}, this._options, options);
      key = this.encodeKey(path, key, options, DEL_OP);
      if (key === false) {
        return false;
      }
      result = AbstractNoSQL.prototype.delSync.call(this, key, options);
      if (result) {
        vKeyPath = options._keyPath;
        options.path = vKeyPath[0];
        options.key = vKeyPath[1];
        delete options._keyPath;
        this.postHooks.trigger(DEL_OP, vKeyPath, [DEL_OP, options]);
      }
      return result;
    };

    SubkeyNoSQL.prototype.delAsync = function(key, options, callback) {
      path = this.getPathArray(options);
      if (isFunction(options)) {
        callback = options;
        options = void 0;
      }
      options = extend({}, this._options, options);
      key = this.encodeKey(path, key, options, DEL_OP);
      if (key === false) {
        return false;
      }
      return AbstractNoSQL.prototype.delAsync.call(this, key, options, (function(_this) {
        return function(err, result) {
          var vKeyPath;
          if (err) {
            return _this.dispatchError(err, callback);
          }
          vKeyPath = options._keyPath;
          options.path = vKeyPath[0];
          options.key = vKeyPath[1];
          delete options._keyPath;
          _this.postHooks.trigger(DEL_OP, vKeyPath, [DEL_OP, options]);
          if (callback) {
            return callback(null, result);
          }
        };
      })(this));
    };

    SubkeyNoSQL.prototype.prepareOperations = function(operations, options) {
      var i, keyEncoding, op, result, vEncoding, vParentPath, valueEncoding;
      keyEncoding = this.keyEncoding(options);
      valueEncoding = this.valueEncoding(options);
      i = 0;
      if (options) {
        vParentPath = options.path;
      }
      while (i < operations.length) {
        op = operations[i];
        if (!op.type) {
          op.type = 'put';
        }
        op.path = this.getPathArray(op, vParentPath);
        if (!op.separator && options && options.separator) {
          op.separator = options.separator;
        }
        result = prepareOperation(this.preHooks, TRANS_OP, op);
        if (result === HALT_OP) {
          delete operations[i];
        } else if (result !== SKIP_OP) {
          op.key = _encodeKey(op.path, op.key, Codec(op.keyEncoding) || keyEncoding, op);
        }
        vEncoding = Codec(op.valueEncoding) || valueEncoding;
        op._keyPath[2] = op.value;
        if (vEncoding) {
          op.value = vEncoding.encode(op.value);
        }
        i++;
      }
      return operations;
    };

    SubkeyNoSQL.prototype.batchSync = function(operations, options) {
      var result;
      if (isArray(operations)) {
        this.prepareOperations(operations, options);
        result = AbstractNoSQL.prototype.batchSync.call(this, operations, options);
        if (result) {
          operations.forEach((function(_this) {
            return function(op) {
              var vKeyPath, vOpType;
              vKeyPath = op._keyPath;
              op.path = vKeyPath[0];
              op.key = vKeyPath[1];
              op.value = vKeyPath[2];
              delete op._keyPath;
              vOpType = op.type === 'del' ? DEL_OP : PUT_OP;
              if (op.triggerAfter !== false) {
                return _this.postHooks.trigger(vOpType, vKeyPath, [vOpType, op]);
              }
            };
          })(this));
        }
        return result;
      } else {
        return AbstractNoSQL.prototype.batchSync.call(this, operations, options);
      }
    };

    SubkeyNoSQL.prototype.batchAsync = function(operations, options, callback) {
      var that;
      if (isFunction(options)) {
        callback = options;
        options = void 0;
      }
      if (isArray(operations)) {
        this.prepareOperations(operations, options);
        that = this;
        AbstractNoSQL.prototype.batchAsync.call(this, operations, options, function(err, result) {
          if (err) {
            return that.dispatchError(err, callback);
          }
          operations.forEach(function(op) {
            var vKeyPath, vOpType;
            vKeyPath = op._keyPath;
            op.path = vKeyPath[0];
            op.key = vKeyPath[1];
            op.value = vKeyPath[2];
            delete op._keyPath;
            vOpType = op.type === 'del' ? DEL_OP : PUT_OP;
            if (op.triggerAfter !== false) {
              return that.postHooks.trigger(vOpType, vKeyPath, [vOpType, op]);
            }
          });
          if (callback) {
            return callback(err, result);
          }
        });
      } else {
        AbstractNoSQL.prototype.batchAsync.call(this, operations, options, callback);
      }
    };

    SubkeyNoSQL.prototype.approximateSizeSync = function(start, end, options) {
      var keyEncoding, vOptions;
      path = this.getPathArray(options);
      keyEncoding = this.keyEncoding();
      options = extend({}, this._options, options);
      vOptions = extend({}, options);
      if (start !== void 0) {
        start = encodeKey(path, start, keyEncoding, options);
      }
      vOptions.path = options.path;
      if (end !== void 0) {
        end = encodeKey(path, end, keyEncoding, options);
      }
      return AbstractNoSQL.prototype.approximateSizeSync.call(this, start, end);
    };

    SubkeyNoSQL.prototype.approximateSizeAsync = function(start, end, callback) {
      var keyEncoding, options, vOptions;
      path = this.getPathArray(options);
      keyEncoding = this.keyEncoding();
      options = extend({}, this._options, options);
      vOptions = extend({}, options);
      if (start !== void 0) {
        start = encodeKey(path, start, keyEncoding, options);
      }
      vOptions.path = options.path;
      if (end !== void 0) {
        end = encodeKey(path, end, keyEncoding, options);
      }
      return AbstractNoSQL.prototype.approximateSizeAsync.call(this, start, end, callback);
    };

    SubkeyNoSQL.prototype.iterator = function(options) {
      path = this.getPathArray(options);
      options = extend({}, this._options, options);
      options.path = path;
      return SubkeyNoSQL.__super__.iterator.call(this, options);
    };

    SubkeyNoSQL.prototype._addHookTo = function(hooks, opType, range, path, callback) {
      if (this._options.path) {
        path = resolvePathArray(this._options.path, path);
        path.shift(0, 1);
      }
      if (isFunction(range)) {
        callback = range;
        range = [path];
      } else if (isString(range)) {
        range = resolveKeyPath(path, range);
      } else if (isObject(range)) {
        range = addPathToRange(path, range);
      } else {
        throw new Error("not implemented yet");
      }
      return hooks.add(opType, range, callback);
    };

    SubkeyNoSQL.prototype.pre = function(opType, range, path, callback) {
      return this._addHookTo(this.preHooks, opType, range, path, callback);
    };

    SubkeyNoSQL.prototype.post = function(opType, range, path, callback) {
      return this._addHookTo(this.postHooks, opType, range, path, callback);
    };

    SubkeyNoSQL.prototype.createSubkey = function(aKeyPath, aOptions, aReadyCallback) {
      if (isFunction(aOptions)) {
        aReadyCallback = aOptions;
        aOptions = void 0;
      }
      aOptions = extend({}, this._options, aOptions);
      return this.SubkeyClass(aKeyPath, aOptions, aReadyCallback);
    };

    SubkeyNoSQL.prototype.createPath = SubkeyNoSQL.prototype.createSubkey;

    SubkeyNoSQL.prototype.subkey = function(aKeyPath, aOptions, aReadyCallback) {
      if (isFunction(aOptions)) {
        aReadyCallback = aOptions;
        aOptions = void 0;
      }
      aOptions = extend({}, this._options, aOptions);
      aOptions.addRef = false;
      return this.SubkeyClass(aKeyPath, aOptions, aReadyCallback);
    };

    SubkeyNoSQL.prototype.path = SubkeyNoSQL.prototype.subkey;

    SubkeyNoSQL.prototype.root = function(aOptions, aReadyCallback) {
      return this.subkey(null, aOptions, aReadyCallback);
    };

    eventable = require('events-ex/ability');

    eventable(SubkeyNoSQL);

    return SubkeyNoSQL;

  })();

}).call(this);

//# sourceMappingURL=nosql-subkey.js.map
