(function() {
  var addPathToRange, compare, compareKeyPath, exports, has, isArray, isArrayLike, isBuffer, isPrimitive, ltgt, minimatch;

  ltgt = require("ltgt");

  minimatch = require("minimatch");

  isArray = Array.isArray;

  isBuffer = Buffer.isBuffer;

  isArrayLike = function(a) {
    return isArray(a) || isBuffer(a);
  };

  isPrimitive = function(a) {
    a = typeof a;
    return "string" === a || "number" === a;
  };

  has = function(o, k) {
    return Object.hasOwnProperty.call(o, k);
  };

  compare = function(a, b) {
    var c, i, l;
    if (isArrayLike(a) && isArrayLike(b)) {
      l = Math.min(a.length, b.length);
      i = 0;
      while (i < l) {
        c = compare(a[i], b[i]);
        if (c) {
          return c;
        }
        i++;
      }
      return a.length - b.length;
    }
    if (isPrimitive(a) && isPrimitive(b)) {
      return (a < b ? -1 : (a > b ? 1 : 0));
    }
    throw new Error("items not comparable:" + JSON.stringify(a) + " " + JSON.stringify(b));
  };

  compareKeyPath = function(aRange, aKeyPath) {
    var l, lastKeyPath, lastRange;
    if (aRange.length > aKeyPath.length) {
      return false;
    }
    l = aRange.length - 1;
    lastRange = aRange[l];
    lastKeyPath = aKeyPath[l];
    if (typeof lastRange !== typeof lastKeyPath) {
      return false;
    }
    if ("string" === typeof lastRange && minimatch(lastKeyPath, lastRange) === false) {
      return false;
    }
    if (aRange.length === 1 && isArrayLike(lastRange)) {
      l++;
    }
    if ((function() {
      var results;
      results = [];
      while (l--) {
        results.push(compare(aRange[l], aKeyPath[l]));
      }
      return results;
    })()) {
      return false;
    }
    return true;
  };

  addPathToRange = function(path, range) {
    var r;
    r = {};
    if (has(range, "lt")) {
      r.lt = [path, range.lt];
    }
    if (has(range, "gt")) {
      r.gt = [path, range.gt];
    }
    if (has(range, "lte")) {
      r.lte = [path, range.lte];
    }
    if (has(range, "gte")) {
      r.gte = [path, range.gte];
    }
    if (has(range, "start")) {
      if (range.reverse) {
        r.lte = [path, range.start];
      } else {
        r.gte = [path, range.start];
      }
    }
    if (has(range, "end")) {
      if (range.reverse) {
        r.gte = [path, range.end];
      } else {
        r.lte = [path, range.end];
      }
    }
    if (has(range, "min")) {
      r.gte = [path, range.min];
    }
    if (has(range, "max")) {
      r.lte = [path, range.max];
    }
    r.reverse = !!range.reverse;
    if (!r.gte && !r.lte) {
      return [path];
    }
    return r;
  };

  exports = module.exports = function(range, key) {
    if (isArrayLike(range)) {
      return compareKeyPath(range, key);
    }
    if (range.lt && compare(key, range.lt) >= 0) {
      return false;
    }
    if (range.lte && compare(key, range.lte) > 0) {
      return false;
    }
    if (range.gt && compare(key, range.gt) <= 0) {
      return false;
    }
    if (range.gte && compare(key, range.gte) < 0) {
      return false;
    }
    return true;
  };

  exports.compare = compare;

  exports.prefix = compareKeyPath;

  exports.addPathToRange = addPathToRange;

}).call(this);

//# sourceMappingURL=range.js.map
