(function() {
  var InvalidArgumentError, LOADING_STATES, LoadingError, NotFoundError, NotImplementedError, NotOpenedError, OBJECT_STATES, PATH_SEP, ReadError, ReadStream, RefObject, SUBKEY_SEP, WriteStream, argumentsAre, assignDeprecatedPrefixOption, codec, defineProperty, deprecate, errors, getPathArray, inherits, isArray, isFunction, isInheritedFrom, isObject, isString, normalizePath, normalizePathArray, path, resolvePath, resolvePathArray, setImmediate, toPath, version;

  defineProperty = require("util-ex/lib/defineProperty");

  isFunction = require("util-ex/lib/is/type/function");

  isString = require("util-ex/lib/is/type/string");

  isObject = require("util-ex/lib/is/type/object");

  isArray = require("util-ex/lib/is/type/array");

  inherits = require("inherits-ex/lib/inherits");

  isInheritedFrom = require("inherits-ex/lib/isInheritedFrom");

  RefObject = require("ref-object/eventable-ref-object");

  try {
    WriteStream = require("nosql-stream/lib/write-stream");
    ReadStream = require('nosql-stream/lib/read-stream');
  } catch (undefined) {}

  codec = require("./codec");

  path = require("./path");

  errors = require("./errors");

  normalizePath = path.normalize;

  normalizePathArray = path.normalizeArray;

  toPath = path.join;

  ReadError = errors.ReadError;

  NotFoundError = errors.NotFoundError;

  InvalidArgumentError = errors.InvalidArgumentError;

  NotImplementedError = errors.NotImplementedError;

  NotOpenedError = errors.NotOpenedError;

  LoadingError = errors.LoadingError;

  setImmediate = global.setImmediate || process.nextTick;

  deprecate = require("depd")("nosql-subkey");

  deprecate.assignProperty = function(object, deprecatedProp, currentProp) {
    if (object[deprecatedProp]) {
      this(deprecatedProp + " property, use `" + currentProp + "` instead.");
      if (!object[currentProp]) {
        object[currentProp] = object[deprecatedProp];
      }
      return delete object[deprecatedProp];
    }
  };

  assignDeprecatedPrefixOption = function(options) {
    return deprecate.assignProperty(options, "prefix", "path");
  };

  PATH_SEP = codec.PATH_SEP;

  SUBKEY_SEP = codec.SUBKEY_SEP;

  getPathArray = codec.getPathArray;

  toPath = path.join;

  resolvePathArray = path.resolveArray;

  resolvePath = path.resolve;

  version = require("../package.json").version;

  OBJECT_STATES = RefObject.prototype.OBJECT_STATES;

  LOADING_STATES = {
    unload: null,
    loading: 0,
    loaded: 1,
    dirtied: 2,
    modifying: 3,
    modified: 4,
    deleted: 5
  };

  argumentsAre = function(args, value, startIndex, endIndex) {
    if (startIndex == null) {
      startIndex = 0;
    }
    if (!(args && args.length)) {
      return true;
    }
    endIndex = endIndex != null ? Math.min(endIndex, args.length - 1) : args.length - 1;
    while (startIndex <= endIndex) {
      if (args[startIndex] !== value) {
        return false;
      }
      ++startIndex;
    }
    return true;
  };

  module.exports = function(dbCore, DefaultReadStream, DefaultWriteStream) {
    var Subkey, cache;
    if (DefaultReadStream == null) {
      DefaultReadStream = ReadStream;
    }
    if (DefaultWriteStream == null) {
      DefaultWriteStream = WriteStream;
    }
    cache = dbCore.cache;
    Subkey = (function() {
      inherits(Subkey, RefObject);

      function Subkey(aKeyPath, aOptions, aCallback) {
        var vKeyPath, vSubkey;
        if (isFunction(aOptions)) {
          aCallback = aOptions;
          aOptions = {};
        }
        if (!(this instanceof Subkey)) {
          vKeyPath = aKeyPath ? normalizePathArray(getPathArray(aKeyPath)) : [];
          vSubkey = cache.createSubkey(toPath(vKeyPath), Subkey.prototype.Class.bind(null, vKeyPath), aOptions, aCallback);
          return vSubkey;
        }
        Subkey.__super__.constructor.call(this, aKeyPath, aOptions, aCallback);
      }

      Subkey.prototype.initialize = function(aKeyPath, aOptions, aReadyCallback) {
        var event, listener, ref, that, vWannaLoad;
        Subkey.__super__.initialize.call(this);
        defineProperty(this, 'db', dbCore);
        defineProperty(this, '_options', aOptions);
        aKeyPath = getPathArray(aKeyPath);
        aKeyPath = aKeyPath ? normalizePathArray(aKeyPath) : [];
        defineProperty(this, '_pathArray', aKeyPath);
        defineProperty(this, 'self', this, {
          writable: false
        });
        defineProperty(this, 'unhooks', []);
        defineProperty(this, 'listeners', {
          ready: this.emit.bind(this, "ready"),
          closing: this.emit.bind(this, "closing"),
          closed: this.emit.bind(this, "closed"),
          error: this.emit.bind(this, "error")
        });
        ref = this.listeners;
        for (event in ref) {
          listener = ref[event];
          dbCore.on(event, listener);
        }
        if (this._initialize) {
          this._initialize(aKeyPath, aOptions);
        }
        defineProperty(this, '_loadingState_', null);
        this.setLoadingState("unload");
        if (aOptions) {
          vWannaLoad = aOptions.load;
        }
        that = this;
        if (vWannaLoad) {
          delete aOptions.load;
          this.load(aReadyCallback);
          return this.on("ready", function() {
            return that.load(aReadyCallback);
          });
        } else if (aReadyCallback) {
          if (dbCore.isOpen() === true) {
            return setImmediate(aReadyCallback.bind(this, null, this));
          } else {
            return this.on("ready", function(err, result) {
              return aReadyCallback.call(that, err, result);
            });
          }
        }
      };

      Subkey.prototype.finalize = function(isFreeSubkeys) {
        var event, i, listener, ref, results, unhooks;
        if (this._finalize) {
          this._finalize(isFreeSubkeys);
        }
        if (isFreeSubkeys !== false) {
          this.freeSubkeys();
        }
        unhooks = this.unhooks;
        i = 0;
        while (i < unhooks.length) {
          unhooks[i]();
          i++;
        }
        this.unhooks = [];
        ref = this.listeners;
        results = [];
        for (event in ref) {
          listener = ref[event];
          results.push(dbCore.removeListener(event, listener));
        }
        return results;
      };

      defineProperty(Subkey.prototype, "sublevels", void 0, {
        get: function() {
          var k, r, result;
          deprecate("sublevels, all subkeys(sublevels) have cached on dbCore.cache now.");
          r = cache.subkeys(toPath(this._pathArray, "*"));
          result = {};
          for (k in r) {
            result[path.basename(k)] = r[k];
          }
          return result;
        }
      });

      defineProperty(Subkey.prototype, "name", void 0, {
        get: function() {
          var l;
          l = this._pathArray.length;
          if (l > 0) {
            return this._pathArray[l - 1];
          } else {
            return PATH_SEP;
          }
        }
      });

      defineProperty(Subkey.prototype, "fullName", void 0, {
        get: function() {
          return PATH_SEP + this._pathArray.join(PATH_SEP);
        }
      });

      defineProperty(Subkey.prototype, "loadingState", void 0, {
        get: function() {
          var vState;
          vState = this._loadingState_;
          if (vState == null) {
            return "unload";
          } else {
            return ["loading", "loaded", "dirtied", "modifying", "modified", "deleted"][vState];
          }
        }
      });

      Subkey.prototype.LOADING_STATES = LOADING_STATES;

      Subkey.prototype.Class = Subkey;

      Subkey.prototype.version = version;

      Subkey.prototype.setLoadingState = function(value, emitted, param1, param2) {
        if (emitted == null) {
          emitted = false;
        }
        this._loadingState_ = LOADING_STATES[value];
        if (emitted) {
          return this.emit(value, this, param1, param2);
        }
      };

      Subkey.prototype.isLoading = function() {
        return this._loadingState_ === LOADING_STATES.loading;
      };

      Subkey.prototype.isLoaded = function() {
        return this._loadingState_ >= LOADING_STATES.loaded;
      };

      Subkey.prototype.isUnload = function() {
        return this._loadingState_ == null;
      };

      Subkey.prototype._loadAsync = function(aCallback) {
        if (this._loadSync) {
          return setImmediate((function(_this) {
            return function() {
              var e, err, error, result;
              try {
                result = _this.loadSync();
              } catch (error) {
                e = error;
                err = e;
                _this.dispatchError(err, aCallback);
                return;
              }
              if (aCallback) {
                return aCallback(null, result);
              }
            };
          })(this));
        } else {
          return setImmediate(aCallback.bind(this, null, this));
        }
      };

      Subkey.prototype.loadAsync = function(aReadyCallback) {
        var err;
        if (this.isUnload() && dbCore.isOpen() === true) {
          this.setLoadingState("loading");
          return this._loadAsync((function(_this) {
            return function(err, result) {
              if (!err) {
                _this.setLoadingState("loaded");
                if (aReadyCallback) {
                  return aReadyCallback(err, result);
                }
              } else {
                _this.setLoadingState("unload");
                return _this.dispatchError(err, callback);
              }
            };
          })(this));
        } else {
          err = dbCore.isOpen() ? new LoadingError('this is already loaded or loading...') : new NotOpenedError();
          return this.dispatchError(err, callback);
        }
      };

      Subkey.prototype.loadSync = function() {
        var err, result;
        result = this.isUnload() && dbCore.isOpen() === true;
        if (result) {
          this.setLoadingState("loading");
          result = this._loadSync ? this._loadSync() : true;
          this.setLoadingState("loaded");
        } else {
          err = dbCore.isOpen() ? new LoadingError('this is already loaded or loading...') : new NotOpenedError();
          throw err;
        }
        return result;
      };

      Subkey.prototype.load = function(aReadyCallback) {
        if (aReadyCallback) {
          return this.loadAsync(aReadyCallback);
        } else {
          return this.loadSync();
        }
      };

      Subkey.prototype.parent = function(options, callback) {
        var createIfMissing, latestParent, p, result, vkeyPath;
        if (!this._pathArray.length) {
          return void 0;
        }
        if (isFunction(options)) {
          callback = options;
          options = {};
        }
        if (options) {
          createIfMissing = options.createIfMissing;
          latestParent = options.latestParent;
          delete options.createIfMissing;
          delete options.latestParent;
        }
        options = this.mergeOpts(options);
        vkeyPath = this._pathArray.slice(0, this._pathArray.length - 1);
        p = toPath(vkeyPath);
        if (createIfMissing === true) {
          return cache.createSubkey(p, Subkey.bind(null, vkeyPath), options, callback);
        } else {
          result = cache.get(p);
          if ((result == null) && latestParent !== true) {
            throw new NotFoundError(p + " path can not be found in cache");
          }
          while ((result == null) && p !== PATH_SEP) {
            p = path.dirname(p);
            result = cache.get(p);
          }
          if (isFunction(callback)) {
            callback(null, result);
          }
          return result;
        }
      };

      Subkey.prototype.setPath = function(aPath, aCallback) {
        var vPath;
        aPath = getPathArray(aPath);
        if (aPath) {
          aPath = normalizePathArray(aPath);
          if (this._pathArray != null) {
            vPath = this.fullName;
          }
          if ((vPath != null) && vPath !== resolvePath(aPath)) {
            cache.del(vPath);
            this.finalize();
            this.initialize(aPath, this._options, aCallback);
            return true;
          }
        }
        if (aCallback) {
          aCallback(new InvalidArgumentError('path argument is invalid'));
        }
        return false;
      };

      Subkey.prototype._addHook = function(key, callback, hooksAdd) {
        var lst, unhook;
        unhook = hooksAdd(key, this._pathArray, callback);
        this.unhooks.push(unhook);
        lst = this.unhooks;
        return function() {
          var i;
          i = lst.indexOf(unhook);
          if (~i) {
            lst.splice(i, 1);
          }
          return unhook();
        };
      };

      Subkey.prototype.mergeOpts = function(opts) {
        var k, o;
        o = {};
        if (this._options) {
          for (k in this._options) {
            if (this._options[k] !== undefined) {
              o[k] = this._options[k];
            }
          }
        }
        if (opts) {
          for (k in opts) {
            if (opts[k] !== undefined) {
              o[k] = opts[k];
            }
          }
        }
        return o;
      };

      Subkey.prototype.isOpen = function() {
        return dbCore.isOpen();
      };

      Subkey.prototype.pathAsArray = function() {
        return this._pathArray.slice();
      };

      Subkey.prototype.prefix = deprecate["function"]((function() {
        return this.pathAsArray();
      }), "prefix(), use `pathAsArray()` instead, or use path() to return string path..");

      Subkey.prototype.path = function(aPath, aOptions, aCallback) {
        if (aPath === undefined) {
          return this.fullName;
        } else {
          return this.subkey(aPath, aOptions, aCallback);
        }
      };

      Subkey.prototype.subkey = function(name, opts, cb) {
        var vKeyPath;
        vKeyPath = resolvePathArray(this._pathArray, name);
        vKeyPath.shift(0, 1);
        if (isFunction(opts)) {
          cb = opts;
          opts = {};
        }
        opts = this.mergeOpts(opts);
        opts.addRef = false;
        return Subkey(vKeyPath, opts, cb);
      };

      Subkey.prototype.createSubkey = function(name, opts, cb) {
        var vKeyPath;
        vKeyPath = resolvePathArray(this._pathArray, name);
        vKeyPath.shift(0, 1);
        if (isFunction(opts)) {
          cb = opts;
          opts = {};
        }
        return Subkey(vKeyPath, this.mergeOpts(opts), cb);
      };

      Subkey.prototype.createPath = Subkey.prototype.createSubkey;

      Subkey.prototype.sublevel = deprecate["function"]((function(name, opts, cb) {
        return this.subkey(name, opts, cb);
      }), "sublevel(), use `subkey(name)` or `path(name)` instead.");

      Subkey.prototype.freeSubkeys = function(aKeyPattern) {
        var k, vSubkeys;
        if (!aKeyPattern) {
          aKeyPattern = toPath(this._pathArray, "*");
        } else {
          aKeyPattern = resolvePath(this._pathArray, aKeyPattern);
        }
        vSubkeys = cache.subkeys(aKeyPattern);
        for (k in vSubkeys) {
          vSubkeys[k].free();
        }
      };

      Subkey.prototype.putSync = function(key, value, opts) {
        if (argumentsAre(arguments, void 0, 1)) {
          value = key;
          key = ".";
        }
        opts = this.mergeOpts(opts);
        assignDeprecatedPrefixOption(opts);
        opts.path = getPathArray(opts.path, this._pathArray);
        return dbCore.putSync(key, value, opts);
      };


      /*
        put it self:
          put(value, cb)
       */

      Subkey.prototype.putAsync = function(key, value, opts, callback) {
        if (argumentsAre(arguments, void 0, 1) || isFunction(value)) {
          callback = value;
          value = key;
          key = ".";
        }
        if (isFunction(opts)) {
          callback = opts;
          opts = void 0;
        }
        opts = this.mergeOpts(opts);
        assignDeprecatedPrefixOption(opts);
        opts.path = getPathArray(opts.path, this._pathArray);
        return dbCore.putAsync(key, value, opts, callback);
      };

      Subkey.prototype.put = function(key, value, opts, cb) {
        if (isFunction(value)) {
          cb = value;
          value = key;
          key = ".";
        } else if (isFunction(opts)) {
          cb = opts;
          opts = void 0;
        }
        if (cb) {
          return this.putAsync(key, value, opts, cb);
        } else {
          return this.putSync(key, value, opts);
        }
      };


      /*TODO: del itself would destroy itself?  see: the post hook itself in init method.
        del itself:
        del(cb)
       */

      Subkey.prototype.delSync = function(key, opts) {
        if (argumentsAre(arguments, void 0)) {
          key = this.path();
        }
        opts = this.mergeOpts(opts);
        assignDeprecatedPrefixOption(opts);
        opts.path = getPathArray(opts.path, this._pathArray);
        return dbCore.delSync(key, opts);
      };

      Subkey.prototype.delAsync = function(key, opts, cb) {
        if (argumentsAre(arguments, void 0) || isFunction(key)) {
          cb = key;
          key = this.path();
          opts = void 0;
        } else if (isFunction(opts)) {
          cb = opts;
          opts = void 0;
        }
        opts = this.mergeOpts(opts);
        assignDeprecatedPrefixOption(opts);
        opts.path = getPathArray(opts.path, this._pathArray);
        return dbCore.delAsync(key, opts, cb);
      };

      Subkey.prototype.del = function(key, opts, cb) {
        if (isFunction(key) || arguments.length === 0) {
          cb = key;
          key = this.path();
        } else if (isFunction(opts)) {
          cb = opts;
          opts = void 0;
        }
        if (cb) {
          return this.delAsync(key, opts, cb);
        } else {
          return this.delSync(key, opts);
        }
      };

      Subkey.prototype.batchSync = function(ops, opts) {
        opts = this.mergeOpts(opts);
        assignDeprecatedPrefixOption(opts);
        opts.path = getPathArray(opts.path, this._pathArray);
        return dbCore.batchSync(ops, opts);
      };

      Subkey.prototype.batchAsync = function(ops, opts, callback) {
        if (isFunction(opts)) {
          callback = opts;
          opts = void 0;
        }
        opts = this.mergeOpts(opts);
        assignDeprecatedPrefixOption(opts);
        opts.path = getPathArray(opts.path, this._pathArray);
        return dbCore.batchAsync(ops, opts, callback);
      };

      Subkey.prototype.batch = function(ops, opts, cb) {
        if (isFunction(opts)) {
          cb = opts;
          opts = void 0;
        }
        if (cb) {
          return this.batchAsync(ops, opts, cb);
        } else {
          return this.batchSync(ops, opts);
        }
      };

      Subkey.prototype.getSync = function(key, opts) {
        if (argumentsAre(arguments, void 0)) {
          opts = key;
          key = ".";
        }
        opts = this.mergeOpts(opts);
        assignDeprecatedPrefixOption(opts);
        opts.path = getPathArray(opts.path, this._pathArray);
        return dbCore.getSync(key, opts);
      };

      Subkey.prototype.getAsync = function(key, opts, cb) {
        var that;
        if (isFunction(opts)) {
          cb = opts;
          opts = void 0;
        } else if (isFunction(key)) {
          cb = key;
          opts = void 0;
          key = ".";
        }
        opts = this.mergeOpts(opts);
        assignDeprecatedPrefixOption(opts);
        opts.path = getPathArray(opts.path, this._pathArray);
        that = this;
        return dbCore.getAsync(key, opts, function(err, value) {
          if (err) {
            return that.dispatchError(err, cb);
          }
          if (cb) {
            return cb.call(that, null, value);
          }
        });
      };

      Subkey.prototype.get = function(key, opts, cb) {
        if (isFunction(opts)) {
          cb = opts;
          opts = void 0;
        } else if (isFunction(key)) {
          cb = key;
          opts = void 0;
          key = ".";
        }
        if (cb) {
          return this.getAsync(key, opts, cb);
        } else {
          return this.getSync(key, opts);
        }
      };

      Subkey.prototype.pre = function(opType, key, hook) {
        return this._addHook(key, hook, dbCore.pre.bind(dbCore, opType));
      };

      Subkey.prototype.post = function(opType, key, hook) {
        return this._addHook(key, hook, dbCore.post.bind(dbCore, opType));
      };

      Subkey.prototype.prepareFindOptions = function(aOptions) {
        var makeData;
        aOptions = this.mergeOpts(aOptions);
        assignDeprecatedPrefixOption(aOptions);
        makeData = aOptions.makeData;
        if (!aOptions.makeData) {
          aOptions.makeData = aOptions.keys !== false && aOptions.values !== false ? function(key, value) {
            return {
              key: key,
              value: value
            };
          } : aOptions.values === false ? function(key) {
            return key;
          } : aOptions.keys === false ? function(_, value) {
            return value;
          } : function() {};
        }
        aOptions.path = getPathArray(aOptions.path, this._pathArray) || this._pathArray;
        return aOptions;
      };

      Subkey.prototype.findSync = function(aOptions) {
        var it, item, makeData, result;
        aOptions = this.prepareFindOptions(aOptions);
        makeData = aOptions.makeData;
        it = dbCore.iterator(aOptions);
        result = [];
        item = it.nextSync();
        while (item !== false) {
          result.push(makeData(item.key, item.value));
          item = it.nextSync();
        }
        return result;
      };

      Subkey.prototype.findAsync = function(aOptions, callback) {
        var it, makeData, nextOne, result;
        if (isFunction(aOptions)) {
          callback = aOptions;
          aOptions = void 0;
        }
        if (!callback) {
          throw new InvalidArgumentError('callback argument required.');
        }
        aOptions = this.prepareFindOptions(aOptions);
        makeData = aOptions.makeData;
        it = dbCore.iterator(aOptions);
        result = [];
        nextOne = function() {
          return it.next(function(err, key, value) {
            if (err) {
              return callback(err, result);
            }
            if (!arguments.length) {
              return callback(err, result);
            }
            result.push(makeData(key, value));
            return nextOne();
          });
        };
        return nextOne();
      };

      Subkey.prototype.find = function(aOptions, callback) {
        if (isFunction(aOptions)) {
          callback = aOptions;
          aOptions = void 0;
        }
        if (callback) {
          return this.findAsync(aOptions, callback);
        } else {
          return this.findSync(aOptions);
        }
      };

      Subkey.prototype.readStream = function(opts) {
        var stream;
        if (!DefaultReadStream) {
          throw new NotImplementedError("please `npm install nosql-stream` to use streamable feature");
        }
        opts = this.mergeOpts(opts);
        assignDeprecatedPrefixOption(opts);
        opts.path = getPathArray(opts.path, this._pathArray) || this._pathArray;
        stream = DefaultReadStream(dbCore, opts);
        return stream;
      };

      Subkey.prototype.createReadStream = Subkey.prototype.readStream;

      Subkey.prototype.valueStream = function(opts) {
        opts = opts || {};
        opts.values = true;
        opts.keys = false;
        return this.readStream(opts);
      };

      Subkey.prototype.createValueStream = Subkey.prototype.valueStream;

      Subkey.prototype.keyStream = function(opts) {
        opts = opts || {};
        opts.values = false;
        opts.keys = true;
        return this.readStream(opts);
      };

      Subkey.prototype.createKeyStream = Subkey.prototype.keyStream;

      Subkey.prototype.writeStream = function(opts) {
        if (!DefaultWriteStream) {
          throw new NotImplementedError("please `npm install nosql-stream` to use streamable feature");
        }
        opts = this.mergeOpts(opts);
        return new DefaultWriteStream(this, opts);
      };

      Subkey.prototype.createWriteStream = Subkey.prototype.writeStream;

      Subkey.prototype.pathStream = function(opts) {
        opts = opts || {};
        opts.separator = PATH_SEP;
        opts.separatorRaw = true;
        opts.gte = "0";
        return this.readStream(opts);
      };

      Subkey.prototype.createPathStream = Subkey.prototype.pathStream;

      return Subkey;

    })();
    return Subkey;
  };

}).call(this);

//# sourceMappingURL=subkey.js.map
